@@ -1,19 +1,19 @@
 /* engine-g13.c - G13 engine.
    Copyright (C) 2000 Werner Koch (dd9jn)
    Copyright (C) 2001, 2002, 2003, 2004, 2005, 2007, 2009 g10 Code GmbH
- 
+
    This file is part of GPGME.
 
    GPGME is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; either version 2.1 of
    the License, or (at your option) any later version.
-   
+
    GPGME is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.
-   
+
    You should have received a copy of the GNU Lesser General Public
    License along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
@@ -25,10 +25,16 @@
 
 #include <stdlib.h>
 #include <string.h>
-#include <sys/types.h>
+#ifdef HAVE_SYS_TYPES_H
+# include <sys/types.h>
+#endif
 #include <assert.h>
-#include <unistd.h>
+#ifdef HAVE_UNISTD_H
+# include <unistd.h>
+#endif
+#ifdef HAVE_LOCALE_H
 #include <locale.h>
+#endif
 #include <fcntl.h> /* FIXME */
 #include <errno.h>
 
@@ -85,7 +91,7 @@ struct engine_g13
 typedef struct engine_g13 *engine_g13_t;
 
 
-static void g13_io_event (void *engine, 
+static void g13_io_event (void *engine,
                             gpgme_event_io_t type, void *type_data);
 
 
@@ -94,7 +100,7 @@ static char *
 g13_get_version (const char *file_name)
 {
   return _gpgme_get_program_version (file_name ? file_name
-				     : _gpgme_get_g13_path ());
+				     : _gpgme_get_default_g13_name ());
 }
 
 
@@ -211,7 +217,7 @@ g13_new (void **engine, const char *file_name, const char *home_dir)
   gpgme_error_t err = 0;
   engine_g13_t g13;
   int argc;
-  char *argv[5];
+  const char *argv[5];
   char *dft_display = NULL;
   char dft_ttyname[64];
   char *dft_ttytype = NULL;
@@ -219,7 +225,7 @@ g13_new (void **engine, const char *file_name, const char *home_dir)
 
   g13 = calloc (1, sizeof *g13);
   if (!g13)
-    return gpg_error_from_errno (errno);
+    return gpg_error_from_syserror ();
 
   g13->status_cb.fd = -1;
   g13->status_cb.dir = 1;
@@ -245,11 +251,11 @@ g13_new (void **engine, const char *file_name, const char *home_dir)
 
 #if USE_DESCRIPTOR_PASSING
   err = assuan_pipe_connect
-    (g13->assuan_ctx, file_name ? file_name : _gpgme_get_g13_path (),
+    (g13->assuan_ctx, file_name ? file_name : _gpgme_get_default_g13_name (),
      argv, NULL, NULL, NULL, ASSUAN_PIPE_CONNECT_FDPASSING);
 #else
   err = assuan_pipe_connect
-    (g13->assuan_ctx, file_name ? file_name : _gpgme_get_g13_path (),
+    (g13->assuan_ctx, file_name ? file_name : _gpgme_get_default_g13_name (),
      argv, NULL, NULL, NULL, 0);
 #endif
   if (err)
@@ -263,7 +269,7 @@ g13_new (void **engine, const char *file_name, const char *home_dir)
       if (asprintf (&optstr, "OPTION display=%s", dft_display) < 0)
         {
 	  free (dft_display);
-	  err = gpg_error_from_errno (errno);
+	  err = gpg_error_from_syserror ();
 	  goto leave;
 	}
       free (dft_display);
@@ -289,7 +295,7 @@ g13_new (void **engine, const char *file_name, const char *home_dir)
 	{
 	  if (asprintf (&optstr, "OPTION ttyname=%s", dft_ttyname) < 0)
 	    {
-	      err = gpg_error_from_errno (errno);
+	      err = gpg_error_from_syserror ();
 	      goto leave;
 	    }
 	  err = assuan_transact (g13->assuan_ctx, optstr, NULL, NULL, NULL,
@@ -306,7 +312,7 @@ g13_new (void **engine, const char *file_name, const char *home_dir)
 	      if (asprintf (&optstr, "OPTION ttytype=%s", dft_ttytype) < 0)
 		{
 		  free (dft_ttytype);
-		  err = gpg_error_from_errno (errno);
+		  err = gpg_error_from_syserror ();
 		  goto leave;
 		}
 	      free (dft_ttytype);
@@ -354,7 +360,10 @@ g13_set_locale (void *engine, int category, const char *value)
   /* FIXME: If value is NULL, we need to reset the option to default.
      But we can't do this.  So we error out here.  G13 needs support
      for this.  */
-  if (category == LC_CTYPE)
+  if (0)
+    ;
+#ifdef LC_CTYPE
+  else if (category == LC_CTYPE)
     {
       catstr = "lc-ctype";
       if (!value && g13->lc_ctype_set)
@@ -362,6 +371,7 @@ g13_set_locale (void *engine, int category, const char *value)
       if (value)
 	g13->lc_ctype_set = 1;
     }
+#endif
 #ifdef LC_MESSAGES
   else if (category == LC_MESSAGES)
     {
@@ -376,11 +386,11 @@ g13_set_locale (void *engine, int category, const char *value)
     return gpg_error (GPG_ERR_INV_VALUE);
 
   /* FIXME: Reset value to default.  */
-  if (!value) 
+  if (!value)
     return 0;
 
   if (asprintf (&optstr, "OPTION %s=%s", catstr, value) < 0)
-    err = gpg_error_from_errno (errno);
+    err = gpg_error_from_syserror ();
   else
     {
       err = assuan_transact (g13->assuan_ctx, optstr, NULL, NULL,
@@ -392,6 +402,7 @@ g13_set_locale (void *engine, int category, const char *value)
 }
 
 
+#if USE_DESCRIPTOR_PASSING
 static gpgme_error_t
 g13_assuan_simple_command (assuan_context_t ctx, char *cmd,
 			   engine_status_handler_t status_fnc,
@@ -442,6 +453,7 @@ g13_assuan_simple_command (assuan_context_t ctx, char *cmd,
 
   return err;
 }
+#endif
 
 
 static gpgme_error_t
@@ -475,7 +487,7 @@ status_handler (void *opaque, int fd)
           TRACE2 (DEBUG_CTX, "gpgme:status_handler", g13,
 		  "fd 0x%x: ERR line: %s",
                   fd, err ? gpg_strerror (err) : "ok");
-	  
+
 	  /* Command execution errors are not fatal, as we use
 	     a session based protocol.  */
 	  data->op_err = err;
@@ -545,7 +557,7 @@ status_handler (void *opaque, int fd)
 	  src = line + 2;
           while (*src == ' ')
             src++;
-	  
+
 	  args = strchr (line + 2, ' ');
 	  if (!args)
 	    args = line + linelen; /* set to an empty string */
@@ -569,12 +581,12 @@ status_handler (void *opaque, int fd)
       else if (linelen >= 7
                && line[0] == 'I' && line[1] == 'N' && line[2] == 'Q'
                && line[3] == 'U' && line[4] == 'I' && line[5] == 'R'
-               && line[6] == 'E' 
+               && line[6] == 'E'
                && (line[7] == '\0' || line[7] == ' '))
         {
           char *src;
 	  char *args;
-	  
+
           for (src=line+7; *src == ' '; src++)
             ;
 
@@ -588,7 +600,7 @@ status_handler (void *opaque, int fd)
             args++;
 
           err = default_inq_cb (g13, src, args);
-          if (!err) 
+          if (!err)
             {
               /* Flush and send END.  */
               err = assuan_send_data (g13->assuan_ctx, NULL, 0);
@@ -602,10 +614,10 @@ status_handler (void *opaque, int fd)
         }
     }
   while (!err && assuan_pending_line (g13->assuan_ctx));
-  
+
   return err;
 }
-  
+
 
 static gpgme_error_t
 add_io_cb (engine_g13_t g13, iocb_data_t *iocbd, gpgme_io_cb_t handler)
@@ -630,16 +642,21 @@ static gpgme_error_t
 start (engine_g13_t g13, const char *command)
 {
   gpgme_error_t err;
+  assuan_fd_t afdlist[5];
   int fdlist[5];
   int nfds;
+  int i;
 
   /* We need to know the fd used by assuan for reads.  We do this by
      using the assumption that the first returned fd from
      assuan_get_active_fds() is always this one.  */
   nfds = assuan_get_active_fds (g13->assuan_ctx, 0 /* read fds */,
-                                fdlist, DIM (fdlist));
+                                afdlist, DIM (afdlist));
   if (nfds < 1)
     return gpg_error (GPG_ERR_GENERAL);	/* FIXME */
+  /* For now... */
+  for (i = 0; i < nfds; i++)
+    fdlist[i] = (int) afdlist[i];
 
   /* We "duplicate" the file descriptor, so we can close it here (we
      can't close fdlist[0], as that is closed by libassuan, and
@@ -740,7 +757,7 @@ g13_io_event (void *engine, gpgme_event_io_t type, void *type_data)
 struct engine_ops _gpgme_engine_ops_g13 =
   {
     /* Static functions.  */
-    _gpgme_get_g13_path,
+    _gpgme_get_default_g13_name,
     NULL,
     g13_get_version,
     g13_get_req_version,
@@ -781,4 +798,6 @@ struct engine_ops _gpgme_engine_ops_g13 =
     g13_io_event,
     g13_cancel,
     g13_cancel_op,
+    NULL,               /* passwd */
+    NULL                /* set_pinentry_mode */
   };
